def get():
    # todo, implement this to return the right connection for the child_class
    # connection = prom.connection.get(child_klass)
    pass

def set():
    # todo, set a connection for a child orm class using interface
#prom.connection.set(
#    orm=klass,
#    interface=Interface,
#    dbname="",
#    username="",
#    password=""
#    host="",
#    port=""
#)
    pass

